#!/usr/bin/env python

from distutils.core import setup

setup(name='ooredis',
      version='1.0',
      author='huangz',
      author_email='huangz1990@gmail.com',
      url='https://github.com/huangz1990/ooredis',
      description='An redis python lib',
      long_description='',
      download_url='',
      packages=['ooredis', 'ooredis.mix',],
      classifiers = [
        'Topic :: Database',
        ],
     )
